package EVolve;

/* EVolve - an Extensible Software Visualization Framework
 * Copyright (C) 2001-2002 Qin Wang
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * EVolve is distributed at http://www.sable.mcgill.ca/EVolve/
 */

import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.TreeMap;
import javax.swing.JFileChooser;
import EVolve.data.DataSource;
import EVolve.data.ElementDefinition;
import EVolve.data.Entity;
import EVolve.data.EntityBuilder;
import EVolve.data.Event;
import EVolve.data.EventBuilder;
import EVolve.data.FieldDefinition;
import EVolve.data.NumericStringComparator;
import EVolve.exceptions.CancelLoadDataException;
import EVolve.exceptions.DataProcessingException;
import EVolve.exceptions.EVolveException;

public class DemoSource implements DataSource {
    private final String dsourceName = "DemoSource";
    private String fn = null; // data file name

    private RandomAccessFile file;

    private EntityBuilder classBuilder;

    private EntityBuilder threadBuilder;

    private EntityBuilder methodBuilder;
    private FieldDefinition methodDefiningClass;

    private EntityBuilder locationBuilder;

    private EntityBuilder sizeBuilder;

    private EventBuilder allocationBuilder;
    private FieldDefinition allocationObjectCount;
    private FieldDefinition allocationObjectSize;
    private FieldDefinition allocationBytecode;
    private FieldDefinition allocationObjectType;
    private FieldDefinition allocationThread;
    private FieldDefinition allocationMethod;
    private FieldDefinition allocationLocation;
    private FieldDefinition allocationSize;
    private FieldDefinition allocationFieldSum;
    private FieldDefinition allocationFieldCounter;

    private EventBuilder invocationBuilder;
    private FieldDefinition invocationNumber;
    private FieldDefinition invocationBytecode;
    private FieldDefinition invocationThread;
    private FieldDefinition invocationMethod;
    private FieldDefinition invocationLocation;
    private FieldDefinition invocationFieldCounter;


    private ElementDefinition[] definition;
    private int definitionCounter;

    private TreeMap classMap;
    private TreeMap threadMap;
    private TreeMap methodMap;
    private TreeMap locationMap;
    private TreeMap sizeMap;

    private long counter[];

    public void init() throws EVolveException {
        String lastName = fn;

        fn = Scene.getDataFileName();
        if (fn == null) {
            JFileChooser fc = new JFileChooser(Scene.getUIManager().getLastDataDir());
            if (fc.showOpenDialog(Scene.getFrame()) == JFileChooser.APPROVE_OPTION) {
                try {
                    file = new RandomAccessFile(fc.getSelectedFile(), "r");
                    fn = fc.getSelectedFile().getName();
                    Scene.setDataFilename(fc.getSelectedFile().getPath());
                    Scene.getUIManager().setLastDataDir(fc.getSelectedFile().getPath());
                    Scene.setDataFilename(null);
                } catch (IOException e) {
                    throw new DataProcessingException("IO exception occurred when access data file.");
                }
            } else {
                fn = lastName;
                throw new CancelLoadDataException();
            }
        } else {
            try {
                file = new RandomAccessFile(fn, "r");
            } catch (IOException e) {
                throw new DataProcessingException("File loading failed.");
            }
        }

    }

    public void startBuildDefinition() throws DataProcessingException {
        String[] propertySum = {"time","sum"};
        String[] propertyCount = {"time","count"};
        String[] propertyAmount = {"amount"};
        String[] propertyCoordinate = {"time","coordinate"};
        String[] propertyThread = {"reference","thread"};

        definition = new ElementDefinition[7];

        classBuilder = new EntityBuilder("Class", "Class");
        definition[0] = classBuilder.buildDefinition();

        threadBuilder = new EntityBuilder("Thread", "Thread");
        definition[1] = threadBuilder.buildDefinition();

        methodBuilder = new EntityBuilder("Method", "Method");
        methodDefiningClass = methodBuilder.buildReferenceDefinition("Defining Class", classBuilder, null, "Defining class of the method");
        definition[2] = methodBuilder.buildDefinition();

        locationBuilder = new EntityBuilder("Location", "Location in Method");
        definition[3] = locationBuilder.buildDefinition();

        sizeBuilder = new EntityBuilder("Size", "Object Size");
        sizeBuilder.addComparator(new NumericStringComparator());
        definition[4] = sizeBuilder.buildDefinition();

        allocationBuilder = new EventBuilder("Object Allocation", "Object allocation event");
        allocationFieldSum = allocationBuilder.buildValueDefinition("Memory Allocated", propertySum, "Size of memory allocated");
        allocationFieldCounter = allocationBuilder.buildValueDefinition("Allocations", propertyCount, "Number of objects allocated");
        allocationObjectCount = allocationBuilder.buildValueDefinition("Object Count", propertyAmount, "Number of objects allocated");
        allocationObjectSize = allocationBuilder.buildValueDefinition("Memory Allocated", propertyAmount, "Size of memory allocated");
        allocationBytecode = allocationBuilder.buildValueDefinition("Bytecode", propertyCoordinate, "Bytecode sequence");
        allocationObjectType = allocationBuilder.buildReferenceDefinition("Object Type", classBuilder, null, "Type of the object");
        allocationThread = allocationBuilder.buildReferenceDefinition("Thread", threadBuilder, propertyThread, "Thread in which the object is allocated");
        allocationMethod = allocationBuilder.buildReferenceDefinition("Allocating Methods", methodBuilder, null, "Method that creates the object");
        allocationLocation = allocationBuilder.buildReferenceDefinition("Allocating Locations", locationBuilder, null, "Location where the object is created");
        allocationSize = allocationBuilder.buildReferenceDefinition("Allocation Size", sizeBuilder, null, "Allocation Size");
        definition[5] = allocationBuilder.buildDefinition();

        invocationBuilder = new EventBuilder("Method Invocation", "Method invocation event");
        invocationFieldCounter = invocationBuilder.buildValueDefinition("Invocations", propertyCount, "Total number of invocations");
        invocationNumber = invocationBuilder.buildValueDefinition("Number of Invocations", propertyAmount, "Total number of invocations");
        invocationBytecode = invocationBuilder.buildValueDefinition("Bytecode", propertyCoordinate, "Bytecode sequence");
        invocationThread = invocationBuilder.buildReferenceDefinition("Thread", threadBuilder, propertyThread, "Thread in which the method is invoked");
        invocationMethod = invocationBuilder.buildReferenceDefinition("Method", methodBuilder, null, "Method that is invoked");
        invocationLocation = invocationBuilder.buildReferenceDefinition("Invoking Locations", locationBuilder, null, "Location where the method is invoked");
        definition[6] = invocationBuilder.buildDefinition();

        definitionCounter = -1;
        counter = new long[definition.length];
        for (int i=0; i<counter.length; i++)
            counter[i] = 0;
    }

    public ElementDefinition getNextDefinition() throws DataProcessingException {
        definitionCounter++;
        if (definitionCounter < definition.length) {
            return definition[definitionCounter];
        } else {
            return null;
        }
    }

    private String getSub(String line, int part) {
        int start = line.indexOf(' ') + 1;
        int i = 1;
        while (i < part) {
            start = line.indexOf(' ', start) + 1;
            i++;
        }
        int end = line.indexOf(' ', start);
        if (end == -1) {
            return line.substring(start);
        } else {
            return line.substring(start, end);
        }
    }

    public void startBuildEntity() throws DataProcessingException {
        try {
            file.seek(0);

            classMap = new TreeMap();
            threadMap = new TreeMap();
            methodMap = new TreeMap();
            locationMap = new TreeMap();
            sizeMap = new TreeMap();
        } catch (IOException e) {
            throw new DataProcessingException("File processing failed.");
        }
    }

    public Entity getNextEntity() throws DataProcessingException {
        try {
            Entity returnVal = null;
            String line = file.readLine();

            while ((returnVal == null) && (line != null) && (line.length()>0)) {
                line = line.trim();
                char ch = line.charAt(0);

                if (ch == 'C') {
                    classBuilder.newEntity(getSub(line, 1));
                    returnVal = classBuilder.buildEntity();
                    classMap.put(getSub(line, 2), returnVal);
                } else if (ch == 'T') {
                    threadBuilder.newEntity(getSub(line, 1));
                    returnVal = threadBuilder.buildEntity();
                    threadMap.put(getSub(line, 2), returnVal);
                } else if (ch == 'M') {
                    methodBuilder.newEntity(getSub(line, 1));
                    methodBuilder.addReferenceField(methodDefiningClass, (Entity)(classMap.get(getSub(line, 3))));
                    returnVal = methodBuilder.buildEntity();
                    methodMap.put(getSub(line, 2), returnVal);
                } else if (ch == 'L') {
                    locationBuilder.newEntity(getSub(line, 1));
                    returnVal = locationBuilder.buildEntity();
                    locationMap.put(getSub(line, 2), returnVal);
                } else if (ch == 'O') {
                    String size = getSub(line, 6);
                    if (!sizeMap.containsKey(size)) {
                        sizeBuilder.newEntity(size);
                        returnVal = sizeBuilder.buildEntity();
                        sizeMap.put(size, returnVal);
                    }
                    counter[5]++;
                } else if (ch == '+') {
                    counter[6]+=2;
                }

                if (returnVal == null) {
                    line = file.readLine();
                }
            }

            return returnVal;
        } catch (IOException e) {
            throw new DataProcessingException("File processing failed.");
        }
    }

    public void startBuildEvent() throws DataProcessingException {
        try {
            file.seek(0);
        } catch (IOException e) {
            throw new DataProcessingException("File processing failed.");
        }
    }

    public Event getNextEvent() throws DataProcessingException {
        try {
            Event returnVal = null;
            String line = file.readLine();

            while ((returnVal == null) && (line != null)) {
                line = line.trim();
                char ch = line.charAt(0);

                if (ch == 'O') {
                    allocationBuilder.newEvent();
                    String size = getSub(line, 6);
                    allocationBuilder.addValueField(allocationObjectCount, 1);
                    allocationBuilder.addValueField(allocationObjectSize, Integer.parseInt(getSub(line, 6)));
                    allocationBuilder.addValueField(allocationFieldCounter, 1);
                    allocationBuilder.addValueField(allocationFieldSum, Integer.parseInt(getSub(line, 6)));
                    allocationBuilder.addValueField(allocationBytecode, Integer.parseInt(getSub(line, 1)));
                    allocationBuilder.addReferenceField(allocationObjectType, (Entity)(classMap.get(getSub(line, 5))));
                    allocationBuilder.addReferenceField(allocationThread, (Entity)(threadMap.get(getSub(line, 2))));
                    allocationBuilder.addReferenceField(allocationMethod, (Entity)(methodMap.get(getSub(line, 3))));
                    allocationBuilder.addReferenceField(allocationLocation, (Entity)(locationMap.get(getSub(line, 4))));
                    allocationBuilder.addReferenceField(allocationSize, (Entity)(sizeMap.get(size)));
                    returnVal = allocationBuilder.buildEvent();
                } else if (ch == '+') {
                    invocationBuilder.newEvent();
                    invocationBuilder.addValueField(invocationNumber, 1);
                    invocationBuilder.addValueField(invocationFieldCounter, 1);
                    invocationBuilder.addValueField(invocationBytecode, Integer.parseInt(getSub(line, 1)));
                    invocationBuilder.addReferenceField(invocationThread, (Entity)(threadMap.get(getSub(line, 2))));
                    invocationBuilder.addReferenceField(invocationMethod, (Entity)(methodMap.get(getSub(line, 3))));
                    invocationBuilder.addReferenceField(invocationLocation, (Entity)(locationMap.get(getSub(line, 4))));
                    returnVal = invocationBuilder.buildEvent();
                } else if (ch == '-') {
                    invocationBuilder.newEvent(true);
                    invocationBuilder.addValueField(invocationNumber, 1);
                    invocationBuilder.addValueField(invocationBytecode, Integer.parseInt(getSub(line, 1)));
                    invocationBuilder.addReferenceField(invocationThread, (Entity)(threadMap.get(getSub(line, 2))));
                    invocationBuilder.addReferenceField(invocationMethod, (Entity)(methodMap.get(getSub(line, 3))));
                    invocationBuilder.addReferenceField(invocationLocation, (Entity)(locationMap.get(getSub(line, 4))));
                    returnVal = invocationBuilder.buildEvent();
                }

                if (returnVal == null) {
                    line = file.readLine();
                }
            }

            return returnVal;
        } catch (IOException e) {
            throw new DataProcessingException("File processing failed.");
        }
    }

    public String getName() {
        return dsourceName;
    }

    public String getFileName() {
        return fn;
    }

    public long getTotalNumberOfEvents() {
        long total = 0;

        for (int i=0; i<definition.length; i++)
            total = total + counter[i];

        return total;
    }

    public long getNumberOfEvents(String definitionName) {
        long number = 0;

        for (int i=0; i<definition.length; i++) {
            if (definition[i].getName().equals(definitionName)) {
                number = counter[i];
                if (definitionName.equals("Method Invocation")) number = number/2;
                break;
            }
        }

        return number;
    }
}