/* EVolve - an Extensible Software Visualization Framework
 * Copyright (C) 2001-2002 Qin Wang
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * EVolve is distributed at http://www.sable.mcgill.ca/EVolve/
 */

package EVolve;

import java.awt.Color;
import java.net.URL;
import java.util.ArrayList;

import javax.swing.JFrame;
import javax.swing.JOptionPane;

import EVolve.data.DataManager;
import EVolve.exceptions.CancelLoadDataException;
import EVolve.exceptions.DataProcessingException;
import EVolve.exceptions.EVolveException;
import EVolve.util.ProgressIndicator;
import EVolve.util.ToolsManager;
import EVolve.util.settings.SceneSetting;
import EVolve.visualization.VisualizationManager;
import EVolve.visualization.VizFactory.VisualizationFactory;

public class Scene {
    public static final String VERSION = "EVolve 1.2.4";
    private static Scene e;

    private DataSourceManager dsm;
    private DataManager dm;
    private VisualizationManager vm;
    private UIManager um;
    private ToolsManager tm;
    private Filter filter;
    private String autoDataFilename;
    private String currentDataFilename;
    private boolean dataLoaded;

    private Scene() {
    }

    public static void start(ArrayList dsList, VisualizationFactory[] factory) {
        e = new Scene();

        e.vm = new VisualizationManager(factory);
        e.dsm = new DataSourceManager(dsList);
        e.um = new UIManager();
        e.tm = new ToolsManager();

        e.autoDataFilename = null;
        e.currentDataFilename = "nothing loaded yet..";

        e.dsm.init();
        e.vm.init();
        e.um.init();
        e.tm.init();
        SceneSetting.v().readDataFromFile();
    }

    public static DataManager getDataManager() {
        return e.dm;
    }

    public static VisualizationManager getVisualizationManager() {
        return e.vm;
    }

    public static DataSourceManager getDataSourceManager() {
        return e.dsm;
    }

    public static UIManager getUIManager() {
        return e.um;
    }

    public static ToolsManager getToolsManager() {
        return e.tm;
    }

    public static Filter getFilter() {
        return e.filter;
    }

    public static JFrame getFrame() {
        return e.um.getFrame();
    }

    public static void setStatus(String status) {
        e.um.setStatus(status);
    }

    public static int getColorRGB() {
        return e.dm.getColorRGB();
    }

    public static Color getColor() {
        return e.dm.getColor();
    }

    public static long getEventCounter() {
        return e.dm.getEventCounter();
    }

    public static Thread loadDataSource() {
        setStatus("Loading data, please wait.");
        e.dataLoaded = e.dm.isDataLoaded();
        e.filter = new Filter();
        Thread thread = new Thread() {
            public void run() {
                try {
                    e.um.disableFileMenus();
                    e.um.disableFunctionMenus();
                    e.dm.init();
                    e.um.init();
                    e.vm.init();
                    setStatus("Data loaded.");
                    e.dm.setDataLoaded(true);
                    e.um.updateDatasourceCombo();
                    e.um.enableFileMenus();
                    e.um.enableMenu();
                } catch (CancelLoadDataException exp) {
                    e.dm.setDataLoaded(e.dataLoaded);
                    if (e.dataLoaded) {
                        e.um.updateDatasourceCombo();
                        e.um.enableMenu();
                    }
                    e.um.enableFileMenus();
                } catch (DataProcessingException exp) {
                    e.dm.setDataLoaded(false);
                    e.um.enableFileMenus();
                    setStatus(exp.getMessage());
                } catch (EVolveException exp) {
                    e.dm.setDataLoaded(false);
                    setStatus(exp.getMessage());//"Invalid data format!");
                    e.um.enableFileMenus();
                } catch (Exception exp) {
                    setStatus("Failed to load trace. Reason: "+exp.getMessage());
                    e.um.enableFileMenus();
                }
            }
        };
        thread.start();
        return thread;
    }

    public static void autoLoadDataSource() {
        try {
            loadDataSource().join();
        } catch (InterruptedException e) {

        }
    }

    public static void visualize() {
        setStatus("Processing data, please wait.");

        ProcessingThread thread = new ProcessingThread();
        ProgressIndicator progress = new ProgressIndicator(getFrame(),e.dm, thread);

        thread.start();
    }

    public static void autoVisualize() {
        setStatus("Processing data, please wait.");

        try {
            e.dm.sendEvents();
            e.vm.visualize();
            setStatus("Visualization finished.");
        } catch (DataProcessingException e) {
            setStatus(e.getMessage());
        }
    }

    public static String getDataFileName() {
        return e.autoDataFilename;
    }

    public static void setDataFilename(String filename) {
        e.autoDataFilename = filename;
        if (filename!=null) e.currentDataFilename = filename;
    }

    public static void clearFilters() {
        e.filter = new Filter();
    }

    public static void showErrorMessage(String msg) {
        JOptionPane.showMessageDialog(e.um.getDesktop(),msg,"Error Message",JOptionPane.ERROR_MESSAGE);
    }

    public static void setCurrentDataFilename(String filename) {
        e.currentDataFilename = filename;
    }

    public static String getCurrentDataFilename() {
        return e.currentDataFilename;
    }

    public static void selectDataSource(DataManager dm, Filter filter) {
        e.dm = dm;
        e.filter = filter;
    }

    public static URL getGifURL(String filename) {
        return e.getClass().getResource("img/"+filename);
    }
}
