/* EVolve - an Extensible Software Visualization Framework
 * Copyright (C) 2001-2002 Qin Wang
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * EVolve is distributed at http://www.sable.mcgill.ca/EVolve/
 */

package EVolve.data;

import java.util.*;

/**
 * Element builder.
 */
public abstract class ElementBuilder {
    private static int typeCounter; // counter of element type

    private String elementName; // name of the elements
    private String elementDescription; // description of the elements

    protected int elementType; // type of the elements
    private ElementDefinition elementDefinition; // definition of the elements

    private ArrayList fieldDefinitionList; // field definitions

    private int entityCounter; // counter of entities
    private String entityName; // name of the entity
    private long[] field; // fields
    private boolean isOptional;

    /**
     * Creates an element builder.
     *
     * @param   elementName name of the elements
     * @param   elementDescription description of the elements
     */
    protected ElementBuilder(String elementName, String elementDescription) {
        this.elementName = elementName;
        this.elementDescription = elementDescription;

        typeCounter++;
        this.elementType = typeCounter;
        this.elementDefinition = null;
        this.fieldDefinitionList = new ArrayList();
        this.entityCounter = -1;
    }

    /**
     * Initializes the element builders.
     */
    public static void init() {
        typeCounter = -1;
    }

    /**
     * Builds the definition of a value field.
     *
     * @param   fieldName name of the field
     * @param   fieldProperty property of the field
     * @param   fieldDescription description of the field
     * @return  the field definition
     */
    public FieldDefinition buildValueDefinition(String fieldName, String[] fieldProperty, String fieldDescription) {
        assert(elementDefinition == null) : ("The definition of element " + elementName + " is already built, cannot add field definition to it.");

        FieldDefinition fieldDefinition = new FieldDefinition(fieldName, fieldDefinitionList.size(), -1, fieldDescription);
        if (fieldProperty != null) {
            for (int i = 0; i < fieldProperty.length; i++) {
                fieldDefinition.addProperty(fieldProperty[i]);
            }
        }
        fieldDefinitionList.add(fieldDefinition);

        return fieldDefinition;
    }

    /**
     * Builds the definition of a reference field.
     *
     * @param   fieldName name of the field
     * @param   referenceBuilder builder of the reference
     * @param   fieldProperty property of the field
     * @param   fieldDescription description of the field
     * @return  the field definition
     */
    public FieldDefinition buildReferenceDefinition(String fieldName, EntityBuilder referenceBuilder, String[] fieldProperty, String fieldDescription) {
        assert (elementDefinition == null) : ("The definition of element " + elementName + " is already built, cannot add field definition to it.");

        FieldDefinition fieldDefinition = new FieldDefinition(fieldName, fieldDefinitionList.size(), referenceBuilder.elementType, fieldDescription);
        if (fieldProperty != null) {
            for (int i = 0; i < fieldProperty.length; i++) {
                fieldDefinition.addProperty(fieldProperty[i]);
            }
        }
        fieldDefinitionList.add(fieldDefinition);

        return fieldDefinition;
    }

    /**
     * Builds the element definition.
     *
     * @return  the element definition
     */
    protected EntityDefinition buildEntityDefinition() {
        assert (elementDefinition == null) : ("The definition of element " + elementName + " is already built.");

        FieldDefinition[] fieldDefinition = new FieldDefinition[fieldDefinitionList.size()];
        for (int i = 0; i < fieldDefinition.length; i++) {
            fieldDefinition[i] = (FieldDefinition)(fieldDefinitionList.get(i));
        }

        elementDefinition = new EntityDefinition(elementName, elementType, fieldDefinition, elementDescription);

        return (EntityDefinition)(elementDefinition);
    }

    /**
     * Builds the element definition.
     *
     * @return  the element definition
     */
    protected EventDefinition buildEventDefinition() {
        assert (elementDefinition == null) : ("The definition of element " + elementName + " is already built.");

        FieldDefinition[] fieldDefinition = new FieldDefinition[fieldDefinitionList.size()];
        for (int i = 0; i < fieldDefinition.length; i++) {
            fieldDefinition[i] = (FieldDefinition)(fieldDefinitionList.get(i));
        }

        elementDefinition = new EventDefinition(elementName, elementType, fieldDefinition, elementDescription);

        return (EventDefinition)(elementDefinition);
    }

    /**
     * Checks if definition is built.
     *
     * @return  true if definition is built
     */
    private boolean checkNewElement() {
        for (int i = 0; i < field.length; i++) {
            field[i] = Integer.MIN_VALUE;
        }
        return (elementDefinition != null);
    }

    /**
     * Starts building new entity.
     *
     * @param   entityName name of the entity
     */
    protected void newEntity(String entityName) {
        field = new long[elementDefinition.getFieldDefinition().length];

        assert (checkNewElement()) : ("The definition of " + elementName + " must be built before building elements.");

        this.entityName = entityName;
    }

    /**
     * Starts building new event.
     */
    protected void newEvent() {
        field = new long[elementDefinition.getFieldDefinition().length];
        isOptional = false;

        assert (checkNewElement()) : ("The definition of " + elementName + " must be built before building elements.");
    }

    /**
     * Starts building new event.
     * @param    isOptional    indicate whether this event is an optional event
     */
    protected void newEvent(boolean isOptional) {
        newEvent();
        this.isOptional = isOptional;

        assert (checkNewElement()) : ("The definition of " + elementName + " must be built before building elements.");
    }

    /**
     * Adds a value field.
     *
     * @param   key key of the field
     * @param   value value of the field
     */
    public void addValueField(FieldDefinition fieldKey, long value) {
        assert (elementDefinition != null) : ("The definition of " + elementName + " must be built before building elements.");
        assert (elementDefinition.getFieldDefinition()[fieldKey.getIndex()].getReference() == -1) : ("Wrong field type, " + fieldKey.getName() + " should be a reference.");
        assert (field[fieldKey.getIndex()] == Integer.MIN_VALUE) : (fieldKey.getName() + "is already added.");

		field[fieldKey.getIndex()] = value;
    }

    /**
     * Adds a reference field.
     *
     * @param   key key of the field
     * @param   reference the reference
     */
    public void addReferenceField(FieldDefinition fieldKey, Entity reference) {
        assert (elementDefinition != null) : ("The definition of " + elementName + " must be built before building elements.");
        assert (elementDefinition.getFieldDefinition()[fieldKey.getIndex()].getReference() != -1) : ("Wrong field type, " + fieldKey.getName() + " should be a value.");
        assert (elementDefinition.getFieldDefinition()[fieldKey.getIndex()].getReference() == reference.getType()) : ("Wrong entity type.");
        assert (field[fieldKey.getIndex()] == Integer.MIN_VALUE) : (fieldKey.getName() + "is already added.");

        field[fieldKey.getIndex()] = reference.getId();
    }

    /**
     * Checks if all fields are added.
     *
     * @return  true if all fields are added
     */
    private boolean checkBuildElement() {
        boolean returnVal = true;
        for (int i = 0; i < field.length -1; i++) {
            if (field[i] == Integer.MIN_VALUE) {
                returnVal = false;
                break;
            }
        }
        return returnVal;
    }

    /**
     * Builds the entity.
     *
     * @return  the entity
     */
    protected Entity buildEntity() {
        assert (elementDefinition != null) : ("The definition of " + elementName + " must be built before building elements.");
        assert (checkBuildElement()) : ("Wrong field number.");

        entityCounter++;
        return new Entity(elementDefinition.getType(), entityCounter, entityName, field);
    }

    /**
     * Builds the event.
     *
     * @return  the event
     */
    protected Event buildEvent() {
        assert (elementDefinition != null) : ("The definition of " + elementName + " must be built before building elements.");
        assert (checkBuildElement()) : ("Wrong field number.");

        return new Event(elementDefinition.getType(), field, isOptional);
    }

    protected Event buildEvent(boolean isOptional) {
        this.isOptional = isOptional;
        return buildEvent();
    }
}