/**
 * Created by IntelliJ IDEA.
 * User: Wei Wang
 * Date: Apr 16, 2003
 * Time: 9:56:12 PM
 */

package EVolve.util.phasedetectors;

import java.awt.Component;
import java.util.ArrayList;

import javax.swing.JLabel;
import javax.swing.JTextField;

import EVolve.Scene;
import EVolve.util.HelperFuncs;
import EVolve.util.equators.Set;
import EVolve.util.equators.UnorderedUnlimitedSet;
import EVolve.util.settings.PhaseDetectorSetting;
import EVolve.visualization.XYViz.XYVisualization;




public class HotspotPhaseDetector extends PhaseDetector{
    private final byte graduateChange = 0x0000, notChange = 0x0001, empty = 0x0002;
    private byte status[];
    private int noiseTolerance, sampleSize;
    private float threshold;
    private final String configName = "hotspot_detector";

    private UnorderedUnlimitedSet workingSet;

    private JTextField[] contents;
    private String[] tags;

    public HotspotPhaseDetector(XYVisualization visual) {
        super(visual);
        workingSet = new UnorderedUnlimitedSet();
        data.add(workingSet);

        threshold = (float)0.8;
        noiseTolerance = 2;
        sampleSize = 1;


        tags = new String[3];
        tags[0] = "[Noise tolerance (intervals)]";
        tags[1] = "[Threshold]";
        tags[2] = "[Sample]";

        JTextField textUpperThreshold, textNoiseTolerance;
        JTextField textSampleSize;
        contents = new JTextField[3];
        contents[0] = textNoiseTolerance = new JTextField();
        contents[0].setColumns(2);

        contents[1] = textUpperThreshold = new JTextField();
        contents[1].setColumns(3);

        contents[2] = textSampleSize = new JTextField();
        contents[2].setColumns(2);

        textUpperThreshold.setText(String.valueOf(threshold));
        textUpperThreshold.setToolTipText("Match threshold, from 0-1");
        textSampleSize.setText(String.valueOf(sampleSize));
        textSampleSize.setToolTipText("Take how many previous intervals as a sample set");
        textNoiseTolerance.setText(String.valueOf(noiseTolerance));
        textNoiseTolerance.setToolTipText("Noise tolerance");
        refreshDetectorParameters();
    }

    public void reset() {
        super.reset();
        workingSet = new UnorderedUnlimitedSet();
        data.add(workingSet);
    }

    public String getName() {
        return "Hotspot Detector";
    }

    protected void refreshDetectorParameters() {
        //PhaseDetectorSetting setting = PhaseDetectorSetting.v();
        //setting.readDataFromFile(tags, contents, configName);
        try {
            noiseTolerance = Integer.parseInt(contents[0].getText());
            threshold = Float.parseFloat(contents[1].getText());
            sampleSize = Integer.parseInt(contents[2].getText());
        } catch (NumberFormatException e) {
            Scene.showErrorMessage("Format of detector configuration file is incorrect.");
            return;
        }
    }

    protected ArrayList autoDetectPhase() {
        ArrayList returnVal = new ArrayList();

        refreshDetectorParameters();

        HelperFuncs.stop = -1;

        if (sampleSize<=0) return returnVal;

        status = new byte[data.size()];
        float[] rates = new float[data.size()];

        for (int iteration = 0; iteration < sampleSize; iteration++) {
            for (int i=iteration; i<data.size(); i=i+sampleSize) {
                if (i<sampleSize) continue;

                float rate = parse(i);
                rates[i] = rate;
                if (rate >= 0) {
                    byte currentStatus = (rate >= threshold) ? notChange : graduateChange;
                    status[i] = currentStatus;
                } else {
                    status[i] = empty;
                }
            }

        }

        int previous = status[0], noise = 1;
        for (int i=1; i<status.length; i++) {
            if (status[i] != previous) {
                noise ++;
                if (noise > noiseTolerance) {
                    if (previous == graduateChange)
                        returnVal.add(new Integer(i-noise));
                    else
                        returnVal.add(new Integer(i-noise+1));
                    noise = 0;
                    previous = status[i];
                }
            } else {
                noise = 0;
            }
        }

        //****************************** logging begins *********************
        System.out.println("Logging ....");
        System.out.println("===============================================");
        for (int i=0; i<status.length/10; i++) {
            System.out.print(""+i*10+"-"+(i*10+9)+":");
            for (int j=0; j<10; j++) {
                System.out.print(status[i*10+j]+",");
            }
            System.out.println("");
            System.out.print(""+i*10+"-"+(i*10+9)+":");
            for (int j=0; j<10; j++) {
                System.out.print(rates[i*10+j]+",");
            }
            System.out.println("");
        }
        System.out.print(""+(status.length/10)*10+"-"+(status.length)+":");
        for (int i=(status.length/10)*10; i<status.length; i++) {
            System.out.print(status[i]+",");
        }
        System.out.println("");
        for (int i=(status.length/10)*10; i<status.length; i++) {
            System.out.print(rates[i]+",");
        }
        System.out.println("");
        //****************************** logging ends *********************

        return returnVal;
    }

    public Component[] createDetectorParamsControls() {
        Component returnVal[] = new Component[tags.length+contents.length];

        for (int i=0; i<tags.length; i++) {
            String prompt = tags[i].substring(1,tags[i].length()-1) + ": ";
            returnVal[i*2] = new JLabel(prompt);
            returnVal[i*2+1] = contents[i];
        }

        return returnVal;
    }

    public void collectData(long xMappedId, long yMappedId) {
        int x = (int)(xMappedId/interval);

        if (x >= data.size()) {
            workingSet = new UnorderedUnlimitedSet();
            while (data.size() < x) {
                data.add(workingSet);
                workingSet = new UnorderedUnlimitedSet();
            }
            data.add(workingSet);
        } /*else {
            workingSet = (UnorderedUnlimitedSet)data.get(x);
        }*/

        workingSet.addElement(yMappedId);
    }

    public void saveSetting() {
        PhaseDetectorSetting.v().save(tags,contents,configName);
    }

    public void triggerPhases(int noiseTolerance, float threshold, boolean append) {
        PhaseEntityTrigger trigger = new PhaseEntityTrigger();
        ArrayList triggered = trigger.gatherTiggeredPhase(data, noiseTolerance, threshold);

        if (triggered.size() == 0) {
            Scene.showErrorMessage("No phase triggered.");
        } else {
            undoRedo.registerAction(triggered,PhaseUndoRedo.triggeredPhases,append);
        }
        undoRedo.update();
    }

    public void entitySetPhases(int noiseTolerance, float threshold, boolean append) {
        PhaseEntitySet entitySet = new PhaseEntitySet();
        ArrayList sets = entitySet.gatherTiggeredPhase(data, noiseTolerance, threshold);

        if (sets.size() == 0) {
            Scene.showErrorMessage("No phase triggered.");
        } else {
            undoRedo.registerAction(sets,PhaseUndoRedo.entitySetPhases,append);
        }
        undoRedo.update();
    }

    private float parse(int checkPoint) {

        float match = 0;

        if (checkPoint == 0) return 0;
        //************** debugging
        if ((HelperFuncs.inDebug)) {
            HelperFuncs.getDebugNo();
        }
        if (checkPoint == HelperFuncs.stop) {
            int qq = 0;
            qq++;
        }
        //************** debugging
        UnorderedUnlimitedSet previous, current;

        current = (UnorderedUnlimitedSet)data.get(checkPoint);
        for (int j=checkPoint+1; j<checkPoint+sampleSize; j++) {
            if (j>=data.size()) break;
            current = (UnorderedUnlimitedSet)current.union((Set)data.get(j));
        }
        previous = (UnorderedUnlimitedSet)data.get(checkPoint - 1);
        for (int j=checkPoint-2; j>=checkPoint-sampleSize; j--) {
            previous = (UnorderedUnlimitedSet)previous.union((Set)data.get(j));
        }
        float unionSize = (float)previous.union(current).size();
        if (unionSize != 0)
            match = ((float)previous.intersection(current).size())/unionSize;
        else
            match = -1;

        return match;
    }

}
