package EVolve.util.settings;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintStream;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;

public class IniFile {
    private Map sections;
    private File file;
    private boolean modified;

    public IniFile(String fileName) throws IOException {
        this(new File(fileName));
    }

    public IniFile(File file) throws IOException {
        this.sections = new LinkedHashMap();
        this.file = file;
        
        if (file.exists()) {
            load();
        }
    }

    private static String parseValue(String value) {
        if (value == null) {
            return null;
        }

        value = value.trim();
        if (value.length() > 0) {
            char s[] = value.toCharArray();
            if (s[0] == '"') {
                // Find the next '"'
                boolean escaped = false;
                int i;
loop:
                for (i = 1; i < s.length; i++) {
                    if (escaped) {
                        escaped = false;
                    } else {
                        switch (s[i]) {
                            case '\\':
                                escaped = true;
                                break;
                            case '"':
                                break loop;
                            default:
                                break;
                        }
                    }
                }

                if (s[i] == '"') {
                    if (i < s.length - 1) {
                        String commentStr = value.substring(i+1).trim();
                        if (!commentStr.startsWith(";")) {
                            throw new RuntimeException("Expression expected to be a comment");
                        }
                    } 
                    value = value.substring(1, i);
                }
            }
        }

        return unescape(value);
    }

    public static String escape(String value) {
        if (value == null) {
            return null;
        }

        String result = "";
        char[] s = value.toCharArray();
        for (int i = 0; i < s.length; i++) {
            char c = s[i];
            switch (c) {
                case '?':
                    result += "\\?";
                    break;
                case '\b':
                    result += "\\b";
                    break;
                case '\t':
                    result += "\\t";
                    break;
                case '\n':
                    result += "\\n";
                    break;
                case '\f':
                    result += "\\f";
                    break;
                case '\r':
                    result += "\\r";
                    break;
                case '"':
                    result += "\\\"";
                    break;
                case '\'':
                    result += "\\'";
                    break;
                case '\\':
                    result += "\\\\";
                    break;
                default:
                    result += c;
                    break;
            }
        }

        return result;
    }

    public static String unescape(String value) {
        if (value == null) {
            return null;
        }

        String result = "";
        char[] s = value.toCharArray();
        for (int i = 0; i < s.length; i++) {
            char c = s[i];
            if (c == '\\') {
                if (i == (s.length - 1)) {
                    return result + '\\';
                } else {
                    c = s[++i];
                    switch (c) {
                        case '?':
                            result += '?';
                            break;
                        case 'b':
                            result += '\b';
                            break;
                        case 't':
                            result += '\t';
                            break;
                        case 'n':
                            result += '\n';
                            break;
                        case 'f':
                            result += '\f';
                            break;
                        case 'r':
                            result += '\r';
                        case 'x':
                            {
                                i++;
                                String hexString = "";
                                while (i < s.length) {
                                    c = s[i];
                                    if (Character.digit(c, 16) >= 0) {
                                        hexString += c;
                                        i++;
                                    } else {
                                        break;
                                    }
                                }

                                if (hexString.length() <= 0) {
                                    i--;
                                    result += "\\x";
                                } else {
                                    result += (char) Short.parseShort(hexString, 16);
                                }
                            }
                            break;
                        case '0':
                        case '1':
                        case '2':
                        case '3':
                        case '4':
                        case '5':
                        case '6':
                        case '7':
                        case '8':
                        case '9':
                            {
                                String octalString = "";
                                while (i < s.length) {
                                    c = s[i];
                                    if (Character.digit(c, 8) >= 0) {
                                        octalString += c;
                                        i++;
                                    } else {
                                        break;
                                    }
                                }

                                if (octalString.length() <= 0) {
                                    result += "\\";
                                } else {
                                    result += (char) Short.parseShort(octalString, 8);
                                }
                            }
                            break;
                        default:
                            result += '\\' + c;
                            break;
                    }
                }
            } else {
                result += c;
            }
        }

        return result;
    }

    public void load() throws IOException {
        BufferedReader in = new BufferedReader(new FileReader(file));
        String line;
        String currentSection = null;

        while ((line = in.readLine()) != null) {
            line = line.trim();
            if (line.length() <= 0) {
                continue;
            }

            char c = line.charAt(0);

            switch (c) {
                case ';':
                    continue;
                case '[':
                    currentSection = line.substring(1, line.length() - 1);
                    put(currentSection);
                    break;
                default:
                    {
                        int eqPos = line.indexOf('=');
                        if (eqPos < 0) {
                            throw new RuntimeException("Invalid INI file format -- Expected key=value");
                        }

                        String key = line.substring(0, eqPos).trim();
                        String value = parseValue(line.substring(eqPos + 1));

                        put(currentSection, key, value);
                    }
                    break;
            }
        }

        modified = false;

        in.close();
    }

    public void write() throws IOException {

        PrintStream out = new PrintStream(new FileOutputStream(file));
        
        Iterator sectionIt = sections.keySet().iterator();
        while (sectionIt.hasNext()) {
            Object sectionName = (String) sectionIt.next();
            Map keysToValues = (Map) sections.get(sectionName);

            out.println("[" + sectionName + "]");
            Iterator keysIt = keysToValues.keySet().iterator();
            while (keysIt.hasNext()) {
                String key = (String) keysIt.next();
                String value = (String) keysToValues.get(key);
                out.println(key + "=\"" + escape(value) + "\"");
            }
            out.println();
        }
    }

    public boolean contains(String section) {
        if (section == null) {
            return false;
        }

        return sections.containsKey(section);
    }

    public boolean contains(String section, String key) {
        if (section == null || key == null) {
            return false;
        }

        if (sections.containsKey(section)) {
            Map keysToValues = (Map) sections.get(section);
            return keysToValues.containsKey(key);
        }

        return false;
    }

    public boolean put(String section) {
        if (section == null) {
            return false;
        }

        if (!sections.containsKey(section)) {
            sections.put(section, new LinkedHashMap());
        }

        modified = true;
        return true;
    }

    public boolean put(String section, String key, String value) {
        if (section == null || key == null) {
            return false;
        }

        if (value == null) {
            value = "";
        }

        Map keysToValues;
        if (sections.containsKey(section)) {
            keysToValues = (Map) sections.get(section);
        } else {
            keysToValues = new LinkedHashMap();
            sections.put(section, keysToValues);
        }

        keysToValues.put(key, value);

        modified = true;
        
        return true;
    }

    public String get(String section, String key) {
        if (sections.containsKey(section)) {
            Map keysToValues = (Map) sections.get(section);
            if (keysToValues.containsKey(key)) {
                return (String) keysToValues.get(key);
            }
        }

        return null;
    }

    public Set getSections() {
        return new HashSet(sections.keySet());
    }

    public Set getKeys(String section) {
        if (section == null) {
            return null;
        }

        if (sections.containsKey(section)) {
            Map keysToValues = (Map) sections.get(section);
            return new HashSet(keysToValues.keySet());
        }

        return null;
    }

    public Set getValues(String section) {
        if (section == null) {
            return null;
        }

        if (sections.containsKey(section)) {
            Map keysToValues = (Map) sections.get(section);
            return new HashSet(keysToValues.values());
        }

        return null;
    }

    public void close() throws IOException {
        if (sections != null) {
            if (modified) {
                this.write();
            }
            sections = null;
        }
    }

    public void finalize() {
        try {
            this.close();
        } catch (IOException e) {
            // not much we can do now...
        }
    }
}

