/* EVolve - an Extensible Software Visualization Framework
 * Copyright (C) 2001-2002 Qin Wang
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * EVolve is distributed at http://www.sable.mcgill.ca/EVolve/
 */

package EVolve.visualization;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.image.BufferedImage;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;

import EVolve.exceptions.NoDataPlotException;

/**
 * Automatic expanding image.
 */
public class AutoImage implements Cloneable{
    protected short[][] image; // the image
    protected int w, h; // width and height of the content
    protected int wMax, hMax; // width and height of the image
    protected int imageWidth, imageHeight;

    protected HashMap object2Int,Int2Object;

    private ArrayList phases;
    private Color gray;

    /**
     * Creates an auto-image.
     */
    public AutoImage() {
        w = 0;
        h = 0;
        wMax = 200;
        hMax = 200;
        image = new short[wMax][hMax];
        for (int i = 0; i < wMax; i++) {
            for (int j = 0; j < hMax; j++) {
                image[i][j] = 0;
            }
        }
        object2Int = new HashMap();
        Int2Object = new HashMap();
        object2Int.put(null,new Integer(0));
        Int2Object.put(new Integer(0),null);
        gray = new Color(234,234,234);
    }

    /**
     * Sets the color of a pixel.
     *
     * @param   x position on X-axis
     * @param   y position on Y-axis
     * @param   color the color
     */
    public void setColor(int x, int y, Object color) {
        if ((x >= wMax) || (y >= hMax)) {
            while (x >= wMax) {
                wMax += 200;
            }
            while (y >= hMax) {
                hMax += 200;
            }

            short[][] newImage = new short[wMax][hMax];
            for (int i = 0; i < wMax; i++) {
                for (int j = 0; j < hMax; j++) {
                    newImage[i][j] = 0;
                }
            }
            for (int i = 0; i < w; i++) {
                for (int j = 0; j < h; j++) {
                    newImage[i][j] = image[i][j];
                }
            }
            image = newImage;
        }

        if (object2Int.containsKey(color)) {
            image[x][y] = ((Integer)object2Int.get(color)).shortValue();
        } else {
            short size = (short)object2Int.size();
            object2Int.put(color,new Integer(size));
            Int2Object.put(new Integer(size),color);
            image[x][y] = size;
        }

        if (x >= w) {
            w = x + 1;
        }
        if (y >= h) {
            h = y + 1;
        }
    }

    /**
     * Gets the color of a pixel.
     *
     * @param   x position on X-axis
     * @param   y position on Y-axis
     * @return  the color
     */
    public Object getColor(int x, int y) {
        if ((image!=null) && ((x < 0) || (y < 0) || (x >= image.length) || (y >= image[0].length))) {
            return null;
        } else {
            int index = image[x][y];
            return Int2Object.get(new Integer(index));
        }
    }

    public Object getSortedColor(ReferenceDimension xAxis, ReferenceDimension yAxis,int x,int y) {
        int xMappedId,yMappedId;
        xMappedId = (xAxis == null) ? x : xAxis.getOriginMappedId(x);
        yMappedId = (yAxis == null) ? y : yAxis.getOriginMappedId(y);

        if ((xMappedId >= wMax) || (yMappedId >= hMax) || (xMappedId < 0) || (yMappedId < 0))
            return null;
        else {
            int index = image[xMappedId][yMappedId];
            if (index == 0)
                return null;
            else
                return (Color)Int2Object.get(new Integer(index));
        }
    }

    /**
     * Gets the buffered-image.
     *
     * @return  buffered-image
     */
    public BufferedImage getImage() throws NoDataPlotException {

        if (w*h == 0) throw new NoDataPlotException();

        BufferedImage returnVal = new BufferedImage(w, h, BufferedImage.TYPE_INT_RGB);
        Graphics2D g2 = (Graphics2D)returnVal.getGraphics();

        g2.setColor(Color.white);
        g2.fillRect(0, 0, w, h);

        if ((phases != null)&&(phases.size()!=0)) {
            int lastX = 0;

            for (int i=0; i<phases.size(); i++) {
                int phaseEnd = ((Integer)phases.get(i)).intValue();

                if ((phaseEnd <= 0)||(phaseEnd>=w)) continue;

                g2.setColor(i%2 == 0 ? gray : Color.white);
                g2.fillRect(lastX,0,phaseEnd,h);
                lastX = phaseEnd;
            }
            g2.setColor(phases.size()%2 == 0 ? gray : Color.white);
            g2.fillRect(lastX,0,w,h);
        }

        for (int i = 0; i < w; i++) {
            for (int j = 0; j < h; j++) {
                if (image[i][j] != 0) {
                    returnVal.setRGB(i, j, ((Color)Int2Object.get(new Integer(image[i][j]))).getRGB());
                }
            }
        }

        return returnVal;
    }

    /**
     * Sorts the image.
     *
     * @param   xAxis dimension that represents the X-axis, null if doesn't need to be sorted
     * @param   yAxis dimension that represents the Y-axis, null if doesn't need to be sorted
     * @return  sorted image
     */
    public AutoImage getSortedImage(ReferenceDimension xAxis, ReferenceDimension yAxis) {
        AutoImage returnVal = new AutoImage();
        for (int i = 0; i < w; i++) {
            for (int j = 0; j < h; j++) {
                int x, y;
                if (xAxis == null) {
                    x = i;
                } else {
                    x = xAxis.getSortedIndex(i);
                }
                if (yAxis == null) {
                    y = j;
                } else {
                    y = yAxis.getSortedIndex(j);
                }
                if ((x != -1) && (y != -1)) {
                    returnVal.setColor(x, y, Int2Object.get(new Integer(image[i][j])));
                }
            }
        }
        return returnVal;
    }

    public int getDimension(int dim) {
        if (dim == -1) return image.length;

        return image[dim].length;
    }

    public int getW() {
        return w;
    }

    public int getH() {
        return h;
    }

    public void setImageSize(int width, int height) {
        imageWidth = w;
        imageHeight = h;
    }

    public short[][] getImageDataArray() {
        return image;
    }

    public void setPhases(ArrayList phases) {
        this.phases = phases;
    }

    public Object clone() {
        AutoImage o = null;
        try {
            o = (AutoImage)super.clone();
        } catch (CloneNotSupportedException e) {
            e.printStackTrace();
            return o;
        }

        o.Int2Object = new HashMap();
        o.object2Int = new HashMap();
        o.object2Int.put(null,new Integer(0));
        o.Int2Object.put(new Integer(0),null);
        o.image = new short[image.length][];
        for (int i=0; i<image.length; i++) {
            o.image[i] = new short[image[i].length];
            for (int j=0; j<image[i].length; j++) {
                o.image[i][j] = image[i][j];
            }
        }
        Iterator it = object2Int.keySet().iterator();
        while (it.hasNext()) {
            Object key = it.next();
            o.object2Int.put(key, object2Int.get(key));
        }
        it = Int2Object.keySet().iterator();
        while (it.hasNext()) {
            Object key = it.next();
            o.Int2Object.put(key, Int2Object.get(key));
        }
        return o;
    }
}