/* EVolve - an Extensible Software Visualization Framework
 * Copyright (C) 2001-2002 Qin Wang
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * EVolve is distributed at http://www.sable.mcgill.ca/EVolve/
 */

package EVolve.visualization;

import java.util.*;
import EVolve.Scene;
import EVolve.data.Element;
import EVolve.data.Entity;
import EVolve.data.EntityComparator;
import EVolve.data.EntityDefinition;
import EVolve.data.LexicalComparator;
import EVolve.data.Selection;
import EVolve.data.ValueComparator;
import EVolve.exceptions.NoDataPlotException;
import EVolve.util.HelperFuncs;
import EVolve.util.Sorter;

public class ReferenceDimension extends Dimension implements Cloneable{
    private ArrayList comparatorList;
    private int currentOrder;
    private int[] order;
    private int [] orderBfUnify;
    private Entity[] entity;
    private HashSet linkedEntity;
    private int selectedIndex;
    private int entityNumber;
    private HashMap Int2Entity, EntityName2Int;
    private Sorter sorter;
    private ArrayList availableSortScheme;
    private boolean sortable;

    public ReferenceDimension() {
        comparatorList = new ArrayList();
        linkedEntity = null;
        selectedIndex = 0;
        entityNumber = -1;
        Int2Entity = new HashMap();
        EntityName2Int = new HashMap();
        orderBfUnify = null;
        availableSortScheme = new ArrayList();
        availableSortScheme.add("Lexical");
        availableSortScheme.add("Temporal");
        sortable = false;
    }

    public ReferenceDimension(boolean sortable) {
        this();
        this.sortable = sortable;
    }

    public int getMaxEntityNumber() {
        return Scene.getDataManager().getEntity()[dataFilter.getTargetType()].size();
    }

    public int getEntityNumber() {
        return entity==null ? -1 : entity.length;
    }

    public long getField(Element element) {
        long returnVal = dataFilter.getData(element);
        Integer mappedId = findMappedId(returnVal);

        if (order[mappedId.intValue()] == Integer.MAX_VALUE) {
            order[mappedId.intValue()] = currentOrder;
            currentOrder++;
        }
        return mappedId.intValue();
    }

    public void preVisualize() {
        currentOrder = 0;
        EntityName2Int.clear();
        Int2Entity.clear();
        order = new int[getMaxEntityNumber()];
        for (int i = 0; i < order.length; i++) {
            order[i] = Integer.MAX_VALUE;
        }
    }

    public void restore() {
        order = new int[getMaxEntityNumber()];
        for (int i = 0; i < order.length; i++) {
            if (i<currentOrder)
                order[i] = i;
            else
                order[i] = Integer.MAX_VALUE;
        }
        linkedEntity = null;

        ArrayList tobeDel = new ArrayList();
        Iterator it = Int2Entity.keySet().iterator();
        while (it.hasNext()) {
            Integer key = (Integer)it.next();
            if (key.intValue() >= currentOrder) {
                tobeDel.add(key);
            }
        }
        for (int i=0; i<tobeDel.size(); i++) {
            Int2Entity.remove(tobeDel.get(i));
        }

        tobeDel.clear();
        it = EntityName2Int.keySet().iterator();
        while (it.hasNext()) {
            Object key = it.next();
            Integer value = (Integer)EntityName2Int.get(key);
            if (value.intValue() >= currentOrder) {
                tobeDel.add(key);
            }
        }
        for (int i=0; i<tobeDel.size(); i++) {
            EntityName2Int.remove(tobeDel.get(i));
        }
    }

    public void visualize() throws NoDataPlotException{

        if (linkedEntity == null) {
            entity = new Entity[currentOrder];

            for (int i = 0; i < order.length; i++) {
                if (order[i] != Integer.MAX_VALUE) {
                    entity[order[i]] = (Entity)Int2Entity.get(new Integer(i));
                }
            }
        } else { // unified reference dimension
            int free = currentOrder;
            entityNumber = currentOrder;
            entity = new Entity[linkedEntity.size()];

            order = new int[linkedEntity.size()];
            Iterator it = linkedEntity.iterator();
            while (it.hasNext()) {
                Entity aEntity = (Entity)it.next();
                Integer mappedId = (Integer)EntityName2Int.get(aEntity.getName());
                if (mappedId != null) {
                    order[mappedId.intValue()] = mappedId.intValue();
                    entity[order[mappedId.intValue()]] = aEntity;
                } else {
                    Int2Entity.put(new Integer(free), aEntity);
                    EntityName2Int.put(aEntity.getName(), new Integer(free));
                    order[free] = Integer.MAX_VALUE;
                    entity[free] = aEntity;
                    free++;
                }
            }
        }

        comparatorList.clear();
        comparatorList.add(new LexicalComparator());
        comparatorList.add(new ValueComparator("Temporal", true, order, EntityName2Int));
        ArrayList entityComparatorList = ((EntityDefinition)(Scene.getDataManager().getElementDefinition()[dataFilter.getTargetType()])).getComparator();
        for (int i = 0; i < entityComparatorList.size(); i++) {
            comparatorList.add(entityComparatorList.get(i));
        }

        if (selectedIndex < comparatorList.size())
            sorter = new Sorter(entity, (EntityComparator)(comparatorList.get(selectedIndex)));

        // we generate an exception on purpose for empty image
        if (entity.length == 0) throw new NoDataPlotException();
    }

    public Entity getEntity(int index) {
        int newMappedId = (sorter == null) ? index : sorter.getSource(index);

        if ((newMappedId < currentOrder)&&(Int2Entity.containsKey(new Integer(newMappedId)))) {
            return (Entity)Int2Entity.get(new Integer(newMappedId));
        }

        return null;
    }

    public ArrayList getComparator() {
        return comparatorList;
    }

    public void selectComparator(int selectedIndex) {
        sorter = new Sorter(entity, (EntityComparator)(comparatorList.get(selectedIndex)));
        this.selectedIndex = selectedIndex;
    }

    public void selectComparator(String comparatorName) {
        for (int i=0; i<comparatorList.size(); i++) {
            if (comparatorName.equals(((EntityComparator)comparatorList.get(i)).getName())) {
                selectComparator(i);
            }
        }
    }

    public void addComparator(EntityComparator comparator) {
        boolean found = false;
        for (int i=0; i<comparatorList.size(); i++) {
            if (((EntityComparator)comparatorList.get(i)).getName().equals(comparator.getName()))
                found = true;
        }
        if (!found) comparatorList.add(comparator);
    }

    public int getSortedIndex(int mappedId) {
        if (order[mappedId] == Integer.MAX_VALUE) {
            return -1;
        } else {
            return sorter.getTarget(order[mappedId]);
        }
    }

    public int getOriginMappedId(int sortedId) {
        int mappedId = sorter.getSource(sortedId);
        return mappedId;
    }

    public void makeSelection(String name, int sourceType, int[] selection) {
        if (selection.length != 0) {
            ArrayList temp = new ArrayList();
            for (int i = 0; i < selection.length; i++) {
                Entity aEntity = getEntity(selection[i]);
                if (aEntity!=null) temp.add(aEntity);
            }

            long[] selected = new long[temp.size()];
            for (int i = 0; i < temp.size(); i++) {
                selected[i] = ((Entity)temp.get(i)).getId();
            }
            Scene.getFilter().addSelection(new Selection(name,sourceType,dataFilter.getTargetType(),dataFilter.getName(),
                                                         selected, 0, Long.MAX_VALUE,null));
        }
    }

    public void makeSelection(String name, int sourceType, int[] selection, long start, long end, ArrayList timeMap) {
        if (selection.length != 0) {
            ArrayList temp = new ArrayList();
            for (int i = 0; i < selection.length; i++) {
                Entity aEntity = getEntity(selection[i]);
                if (aEntity!=null) temp.add(aEntity);
            }

            long[] selected = new long[temp.size()];
            for (int i = 0; i < temp.size(); i++) {
                selected[i] = ((Entity)temp.get(i)).getId();
            }
            Scene.getFilter().addSelection(new Selection(name, sourceType, dataFilter.getTargetType(),
                                                         dataFilter.getName(),selected, start, end, timeMap));
        }
    }

    public Entity[] getEntities() {
        return entity;
    }

    public void linkEntities(HashSet linkedEntity) {
        if (this.linkedEntity != null)
            restore();
        this.linkedEntity = linkedEntity;
    }

    public void clearEntityMap() {
        linkedEntity = null;
        entity = null;
        entityNumber = -1;
    }

    public int getSelectedComparatorIndex() {
        return selectedIndex;
    }

    public String getSelectedComparatorName() {
        if (comparatorList.size() == 0) {
            return new LexicalComparator().getName();
        }
        return ((EntityComparator)comparatorList.get(selectedIndex)).getName();
    }

    public int[] getOrdering() {
        if (orderBfUnify == null)
            return order;
        else
            return orderBfUnify;
    }

    public void updateOrdering(int[] newOrder) {
        if (orderBfUnify == null)
            orderBfUnify = order;
        order = newOrder;
    }

    public int getEntityNumberBeforeLink() {
        return entityNumber;
    }

    public HashMap getEntityName2IntMap() {
        return EntityName2Int;
    }

    public HashMap getInt2EntityMap() {
        return Int2Entity;
    }

    public Entity getEntityFromInt(int mappedId) {
        Integer key = new Integer(mappedId);
        if (Int2Entity.containsKey(key)) {
            return (Entity)Int2Entity.get(key);
        } else
            return null;
    }

    private Integer findMappedId(long entityId) {
        Long value = new Long(entityId);
        Entity entity = (Entity)Scene.getDataManager().getEntity()[dataFilter.getTargetType()].get(value);

        if (EntityName2Int.containsKey(entity.getName())) {
            return (Integer)EntityName2Int.get(entity.getName());
        } else {
            Integer mappedId = new Integer(Int2Entity.size());
            Int2Entity.put(mappedId,entity);
            EntityName2Int.put(entity.getName(),mappedId);
            return mappedId;
        }
    }

    public ArrayList getAvailableSortSchemes() {

        ArrayList entityComparatorList = ((EntityDefinition)(Scene.getDataManager().getElementDefinition()[dataFilter.getTargetType()])).getComparator();

        ArrayList returnVal = new ArrayList();
        for (int i=0; i<availableSortScheme.size(); i++) {
            returnVal.add(availableSortScheme.get(i));
        }
        for (int i = 0; i < entityComparatorList.size(); i++) {
            String newScheme = ((EntityComparator)entityComparatorList.get(i)).getName();
            if (!availableSortScheme.contains(newScheme))
                returnVal.add(newScheme);
        }

        return returnVal;
    }

    public void addSortSchemes(String newScheme) {
        if (!availableSortScheme.contains(newScheme))
            availableSortScheme.add(newScheme);
    }

    public boolean isSortable() {
        return sortable;
    }

    public Object clone() {
        ReferenceDimension o = (ReferenceDimension) super.clone();
        o.comparatorList = (ArrayList)comparatorList.clone();
        o.comparatorList.clear();
        for (int i=0; i<comparatorList.size(); i++) {
            o.comparatorList.add(((EntityComparator)comparatorList.get(i)).clone());
        }
        o.entity = null;
        if (entity != null) {
            o.entity = new Entity[entity.length];
            for (int i=0; i<o.entity.length; i++)
                o.entity[i] = entity[i];
        }
        o.linkedEntity = HelperFuncs.cloneHashSet(linkedEntity);
        o.Int2Entity = HelperFuncs.cloneHashMap(Int2Entity);
        o.EntityName2Int = HelperFuncs.cloneHashMap(EntityName2Int);

        o.order = null;
        if (order != null) {
            o.order = new int[order.length];
            for (int i=0; i<order.length; i++) {
                o.order[i] = order[i];
            }
        }

        o.sorter = (sorter == null) ? null : (Sorter)sorter.clone();
        o.sortable = sortable;
        return o;
    }
}