/* EVolve - an Extensible Software Visualization Framework
 * Copyright (C) 2001-2002 Qin Wang
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * EVolve is distributed at http://www.sable.mcgill.ca/EVolve/
 */

package EVolve.visualization;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import EVolve.Scene;
import EVolve.data.ElementDefinition;
import EVolve.exceptions.NoDataPlotException;
import EVolve.util.phasedetectors.PhaseOperation;
import EVolve.util.xmlutils.datastructures.SerializedVisualization;
import EVolve.visualization.VizFactory.VisualizationFactory;
import EVolve.visualization.XYViz.XYVisualization;

public class VisualizationManager {
    private VisualizationFactory[] factory;
    private ArrayList visualizationList;
    private ArrayList toVisualize;
    private Visualization[][] visualization;

    private JDialog dialog;
    private JList listVisualization;
    private DefaultListModel model;

    private boolean phaseDetectorEnabled;
    private PhaseOperation phaseOperation;

    public VisualizationManager(VisualizationFactory[] factory) {
        this.factory = factory;
        this.visualizationList = new ArrayList();
        this.toVisualize = new ArrayList();
        this.dialog = null;
        phaseDetectorEnabled = false;
        phaseOperation = new PhaseOperation();
    }

    public void init() {
        ArrayList newVisualizationList = new ArrayList();
        ArrayList newToVisualize = new ArrayList();

        for (int i=0; i<visualizationList.size(); i++) {
            Visualization visual = (Visualization)visualizationList.get(i);
            if (visual.getDataSourceId() != Scene.getDataSourceManager().getCurrentDataSourceId())
                newVisualizationList.add(visual);
        }

        for (int i=0; i<toVisualize.size(); i++) {
            Visualization visual = (Visualization)toVisualize.get(i);
            if (visual.getDataSourceId() != Scene.getDataSourceManager().getCurrentDataSourceId())
                newToVisualize.add(visual);
        }
        visualizationList = newVisualizationList;
        toVisualize = newToVisualize;
    }

    private void createDialog() {
        dialog = new JDialog(Scene.getFrame(), "Select Visualizations", true);
        model = new DefaultListModel();
        listVisualization = new JList(model);
        listVisualization.setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);

        JScrollPane scrollPane = new JScrollPane(listVisualization);
        scrollPane.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), "Choose visualizations"));
        dialog.getContentPane().add(scrollPane, BorderLayout.CENTER);

        JPanel panelButton = new JPanel(new FlowLayout());
        dialog.getContentPane().add(panelButton, BorderLayout.SOUTH);

        JButton buttonApply = new JButton("Apply");
        buttonApply.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                dialogApply();
            }
        });
        panelButton.add(buttonApply);

        JButton buttonCancel = new JButton("Cancel");
        buttonCancel.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                dialogCancel();
            }
        });
        panelButton.add(buttonCancel);
    }

    public String[] getFactoryName() {
        ArrayList list = new ArrayList();
        for (int i = 0; i < factory.length; i++) {
            ElementDefinition[] temp = Scene.getDataManager().getElementDefinition(factory[i].getDefinition());
            if (temp.length != 0) {
                list.add(factory[i]);
            }
        }

        String[] returnVal = new String[list.size()];
        for (int i = 0; i < returnVal.length; i++) {
            returnVal[i] = ((VisualizationFactory)(list.get(i))).getName();
        }
        return returnVal;
    }

    public Visualization newVisualization(String name) {
        Visualization newVisualization = null;

        for (int i = 0; i < factory.length; i++) {
            if (name.equals(factory[i].getName())) {
                newVisualization = factory[i].createVisualization();
                newVisualization.setName(name);
                newVisualization.setDataSourceId(Scene.getDataSourceManager().getCurrentDataSourceId());
                newVisualization.setFactory(factory[i]);
                newVisualization.setDefinition(factory[i].getDefinition());
                break;
            }
        }
        visualizationList.add(newVisualization);
        toVisualize.add(newVisualization);
        Scene.getUIManager().addVisualization(newVisualization);
        return newVisualization;
    }

    public void removeVisualization(Visualization visualization) {
        visualizationList.remove(visualization);
        toVisualize.remove(visualization);
    }

    public void preVisualize() {
        if (dialog == null) {
            createDialog();
        }

        int[] selected = new int[toVisualize.size()];
        int selectedCounter = 0;

        model.removeAllElements();

        for (int i = 0; i < visualizationList.size(); i++) {
            model.addElement(((Visualization)(visualizationList.get(i))).getName());
            if (toVisualize.indexOf(visualizationList.get(i)) != -1) {
                selected[selectedCounter] = i;
                selectedCounter++;
            }
        }

        listVisualization.setSelectedIndices(selected);

        Scene.getUIManager().showDialog(dialog, 270, 360);
    }

    public void receiveEvent(EVolve.data.Event event) {
        int currentDsourceId = Scene.getDataSourceManager().getCurrentDataSourceId();
        for (int i = 0; i < visualization[event.getType()].length; i++) {
            int dataSourceId = visualization[event.getType()][i].getDataSourceId();
            if (dataSourceId == currentDsourceId)
                visualization[event.getType()][i].receiveElement(event);
        }
    }

    public void visualize() {
        try {
            for (int i = 0; i < visualization.length; i++) {
                for (int j = 0; j < visualization[i].length; j++) {
                    int dataSourceId = visualization[i][j].getDataSourceId();
                    if (dataSourceId != Scene.getDataSourceManager().getCurrentDataSourceId())
                            continue;

                    for (int k = 0; k < visualization[i][j].getDimension().length; k++) {
                        if (visualization[i][j].getDimension()[k] instanceof ReferenceDimension) {
                            ((ReferenceDimension)(visualization[i][j].getDimension()[k])).visualize();
                        }
                    }
                    visualization[i][j].visualize();
                    visualization[i][j].updateMenu();
                    visualization[i][j].enableSortMenu();
                }
            }
        } catch (NoDataPlotException e) { //a non-fatal exception, show message and then ignore it
            Scene.showErrorMessage(e.getMessage());
        }

    }

    private void dialogApply() {
        dialog.setVisible(false);

        toVisualize.clear();

        int[] selected = listVisualization.getSelectedIndices();

        for (int i = 0; i < selected.length; i++) {
            toVisualize.add(visualizationList.get(selected[i]));
        }

        if (toVisualize.size() != 0) {
            prepareListToBeVisualized();
            Scene.visualize();
        }
    }

    private void dialogCancel() {
        dialog.setVisible(false);
    }

    public ArrayList getVisualizationType() {
        ArrayList returnVal = new ArrayList();
        for (int i = 0; i < visualizationList.size(); i++) {
            String s = "" + ((Visualization)(visualizationList.get(i))).getSubjectDefinition().getType() + "";
            if (returnVal.indexOf(s) == -1) {
                returnVal.add(s);
            }
        }

        return returnVal;
    }

    public VisualizationFactory getFactory(String name) {
        VisualizationFactory retFac = null;
        for (int i = 0; i < factory.length; i++) {
            if (name.equals(factory[i].getName())) {
                retFac = factory[i];
                break;
            }
        }
        return retFac;
    }

    public void autoPreVisualize(SerializedVisualization conf) {
        Visualization visual= (Visualization)visualizationList.get(0);

        visual.restoreConfiguration(conf);

        toVisualize.clear();

        toVisualize.add(visualizationList.get(0));

        if (toVisualize.size() != 0) {
            prepareListToBeVisualized();
        }
    }

    public void autoSaveResult(String path, String dataFn) {
        Visualization visual= (Visualization)visualizationList.get(0);
        visual.autoSave(path,dataFn);
    }

    public ArrayList getVisualizationList() {
        return visualizationList;
    }

    public ArrayList getVisualizationList(int dataSourceId) {
        ArrayList list = new ArrayList();
        int dsourceId = Scene.getDataSourceManager().getCurrentDataSourceId();

        for (int i=0; i<visualizationList.size(); i++) {
            Visualization visual = (Visualization)visualizationList.get(i);
            if (visual.getDataSourceId() == dsourceId)
                list.add(visual);
        }
        return list;
    }

    public void prepareListToBeVisualized() {
        visualization = new Visualization[Scene.getDataManager().getElementDefinition().length][];
        for (int i = 0; i < visualization.length; i++) {
            ArrayList list = new ArrayList();
            for (int j = 0; j < toVisualize.size(); j++) {
                if (((Visualization)(toVisualize.get(j))).getSubjectDefinition().getType() == i) {
                    list.add(toVisualize.get(j));
                }
            }
            visualization[i] = new Visualization[list.size()];
            for (int j = 0; j < visualization[i].length; j++) {
                visualization[i][j] = (Visualization)(list.get(j));
                visualization[i][j].preVisualize();
                for (int k = 0; k < visualization[i][j].getDimension().length; k++) {
                    visualization[i][j].getDimension()[k].preVisualize();
                }
            }
        }
    }

    // following methods are used in predefined visualization
    public void cleanup() {
        Scene.getUIManager().clearWindowsList();
        for (int i=0; i<visualizationList.size(); i++) {
            Visualization visual =(Visualization) visualizationList.get(i);
            visual.getWindow().dispose();
        }
        init();
    }

    // following methods are used in linked visualization
    public void addAllVisualizations() {
        toVisualize.clear();
        for (int i=0; i<visualizationList.size(); i++)
            toVisualize.add(visualizationList.get(i));
    }

    public void prepareForLinkedViz(ArrayList viz,int beginRepaint) {
        toVisualize.clear();

        for (int i=beginRepaint; i< viz.size(); i++) {
            toVisualize.add(viz.get(i));
        }
        if (toVisualize.size() != 0)
            prepareListToBeVisualized();
    }

    public Visualization prepareForAsynOverlapViz(String name) {
        Visualization newVisualization = null;

        for (int i = 0; i < factory.length; i++) {
            if (name.equals(factory[i].getName())) {
                newVisualization = factory[i].createVisualization();
                newVisualization.setName(name);
                newVisualization.setFactory(factory[i]);
                newVisualization.setDefinition(factory[i].getDefinition());
                break;
            }
        }
        return newVisualization;
    }

    public void visualizeNow() {
        Visualization visual = getActiveVisualization();

        if (visual != null) {
            int dataSourceId = visual.getDataSourceId();

            if (dataSourceId != Scene.getDataSourceManager().getCurrentDataSourceId()) {
                Scene.getUIManager().switchDataSource(dataSourceId);
            }

            visualization = new Visualization[Scene.getDataManager().getElementDefinition().length][];
            for (int i = 0; i < visualization.length; i++) {
                ArrayList list = new ArrayList();
                if (visual.getSubjectDefinition().getType() == i) {
                    list.add(visual);
                }
                visualization[i] = new Visualization[list.size()];
                for (int j = 0; j < visualization[i].length; j++) {
                    visualization[i][j] = (Visualization)(list.get(j));
                    visualization[i][j].preVisualize();
                    for (int l = 0; l < visualization[i][j].getDimension().length; l++) {
                        (visualization[i][j].getDimension()[l]).preVisualize();
                    }
                }
            }

            Scene.visualize();
        }
   }

    public void cloneVisualization() {
        Visualization visual = getActiveVisualization();

        if (visual == null) return;

        SerializedVisualization configuration = visual.getCurrentConfiguration();

        Visualization newViz = (Visualization)visual.clone();

        visualizationList.add(newViz);
        toVisualize.add(newViz);
        Scene.getUIManager().addVisualization(newViz);
        configuration.WindowTitle += " - cloned";
        newViz.setName(visual.getName()+" - cloned");
        newViz.restoreConfiguration(configuration);

        try {
            newViz.getWindow().setBounds(visual.getWindow().getBounds());
            newViz.visualize();
            newViz.updateMenu();
            newViz.enableSortMenu();
        } catch (Exception e) {
            // when the clone object is not visualized yet, this
            // harmless exception happens.
        }
    }

    public Visualization getActiveVisualization() {

        Visualization visual = null;

        for (int i=0; i<visualizationList.size(); i++) {
            visual = (Visualization)visualizationList.get(i);
            if (visual.getWindow().isSelected()) {
                return visual;
            }
            visual = null;
        }

        return visual;
    }

    public void enablePhaseDetector(boolean enabled) {
        phaseDetectorEnabled = enabled;
        for (int i=0; i<visualizationList.size(); i++) {
            Visualization visual = (Visualization)visualizationList.get(i);

            if (visual instanceof TableViz) continue;

            ((XYVisualization)visual).enablePhaseDetector(enabled);
        }
    }

    public boolean isPhaseDetectorEnabled() {
        return phaseDetectorEnabled;
    }

    public void refreshPhases() {
        for (int i=0; i<visualizationList.size(); i++) {
            Visualization visual = (Visualization)visualizationList.get(i);
            visual.sort();
        }

    }

    public PhaseOperation getPhaseOperation() {
        return phaseOperation;
    }

    public ArrayList removeDataSource(int[] tobeRemoved) {
        ArrayList dmList = Scene.getDataSourceManager().getDataManagerList();
        HashMap idMap = new HashMap();

        int k=0;
        for (int i=0; i<dmList.size(); i++) {
            boolean remove = false;
            for (int j=0; j<tobeRemoved.length; j++) {
                if (i==tobeRemoved[j]) {
                    remove = true;
                    k++;
                }
            }
            if (!remove)
                idMap.put(new Integer(i),new Integer(i-k));
        }

        ArrayList newVisualizationList = new ArrayList();
        ArrayList newWindowList = new ArrayList();
        for (int i=0; i<visualizationList.size(); i++) {
            Visualization visual = (Visualization)visualizationList.get(i);
            int oldDatasourceId = visual.getDataSourceId();
            if (!idMap.containsKey(new Integer(oldDatasourceId))) {
                EVolve.Window window = visual.getWindow();
                window.dispose();
            } else {
                visual.setDataSourceId(((Integer)idMap.get(new Integer(oldDatasourceId))).intValue());
                newWindowList.add(visual.getWindow());
                newVisualizationList.add(visual);
            }
        }
        visualizationList = newVisualizationList;

        return newWindowList;
    }
}

